import numpy as np
cimport numpy as np
DTYPE = np.float
DTYPE_int = np.int

cdef SimplexPr(np.ndarray X):
    cdef int C = X.shape[0]
    cdef int N = X.shape[1]
    #to sort in descending order
    X = np.array(X).astype(DTYPE)
    cdef np.ndarray T = np.sort(X,axis=0) [::-1]
    cdef np.ndarray t = np.zeros([C,1], dtype=DTYPE) 
    cdef int i, j, kk, k
    cdef float theta, tep, sum_t
    for i in range(N):
        sum_t = 0
        kk = -1
        t = T[:,i]
        #find kk where first kk element >= 1 otherwise kk = last index
        for j in range(C):
            tep = 0
            for k in range(j+1):
                tep = tep + t[k]
            tep = t[j] - (tep-1)/(j+1)
            #tep = t[j] - (np.sum(t[0:j+1]) - 1)/(j+1)
            if tep <= 0:
                kk = j-1
                break
        if kk == -1:
            kk = C-1
        #scale X to be at 1
        for k in range(kk+1):
            sum_t = sum_t+t[k]
        theta = (sum_t-1)/(kk+1)
        #theta = (np.sum(t[0:kk+1]) - 1)/(kk+1)
        X[:,i] = (X[:,i] - theta).clip(min=0).flatten()
    return X

def LAE (np.ndarray x, np.ndarray U, int cn):
    cdef int d = U.shape[0]
    cdef int s = U.shape[1]
    cdef np.ndarray z0 = np.ones((s,1), dtype=DTYPE)/s
    cdef np.ndarray z1 = z0
    cdef np.ndarray delta = np.zeros((1,cn+2), dtype=DTYPE)
    delta[0][0] = 0
    delta[0][1] = 1
    cdef np.ndarray beta = np.zeros((1,cn+1), dtype=DTYPE)
    beta[0][0] = 1
    cdef int t, j
    cdef float alpha, b
    cdef np.ndarray v = np.zeros([s,1], dtype=DTYPE)
    cdef np.ndarray dif = np.zeros([d,1], dtype=DTYPE)
    cdef np.ndarray gv = np.zeros([1,1], dtype=DTYPE)
    cdef np.ndarray dgv = np.zeros([s,1], dtype=DTYPE)
    cdef np.ndarray z = np.zeros([s,1], dtype=DTYPE)
    cdef np.ndarray gz = np.zeros([1,1], dtype=DTYPE)
    cdef np.ndarray gvz = np.zeros([1,1], dtype=DTYPE)
    for t in range(cn):
        alpha = (delta[0][t]-1)/delta[0][t+1]
        v = z1 + alpha*(z1-z0)
        dif = x - np.matmul(U,v)
        gv = np.matmul(dif.transpose(),dif/2)
        dgv = np.matmul(U.transpose(),np.matmul(U,v)-x)
        for j in range(d+1):
            b = 2**j*beta[0][t]
            z = SimplexPr(v-dgv/b)
            dif = x - np.matmul(U,z)
            gz = np.matmul(dif.transpose(),dif/2)
            dif = z - v
            gvz = gv + np.matmul(dgv.transpose(),dif) + b * np.matmul(dif.transpose(),dif/2)
            if gz <= gvz:
                beta[0][t+1] = b
                z0 = z1
                z1 = z
                break
        if beta[0][t+1] == 0:
            beta[0][t+1] = b
            z0 = z1
            z1 = z
        delta[0][t+2] = (1+np.sqrt(1+4*delta[0][t+1]**2))/2
        if np.sum(abs(z1-z0)) <= 1e-4:
            break
    z = z1
    return z, np.sum(abs(z1-z0))

def find_regression(np.ndarray TrainData, np.ndarray val, np.ndarray pos, Anchor, int s, int cn):
    cdef int i, D, n
    cdef float diff
    D = TrainData.shape[0]
    n = TrainData.shape[1]
    cdef np.ndarray x = np.zeros([D], dtype=DTYPE)
    cdef np.ndarray U = np.zeros([D,s], dtype=DTYPE)
    for i in range(n):
        x = TrainData[:,i]
        x = (x/np.linalg.norm(x,2)).reshape((len(x),1))
        U = Anchor[:,pos[i,:]]
        a = LAE(x,U,cn)
        val[i,:] = a[0].flatten()
        diff = diff + a[1]
    return val, diff



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

def calculate_marker_gene(data, labels, topn, gene_labels):
    #TODO: add labels for the genes
    #data gene expression matrix gene x cell
    num_gene, num_cell = data.shape
    num_cluster = len(np.unique(labels))
    cluster_order = []
    gene_idxv = []
    num_cells_inC = []

    #calculate mean of gene expression for each cluster
    #gene x cluster
    gene_mean = np.zeros((num_gene,num_cluster))
    gene_DE_score = np.zeros((num_gene,num_cluster))
    for i in range(num_cluster):
        ind = np.where(labels.flatten() == np.unique(labels)[i])
        gene_mean[:,i] = np.mean(data[:,ind[0]], axis=1).flatten()
        cluster_order = np.append(cluster_order,ind).astype(int)
        num_cells_inC = np.append(num_cells_inC, len(ind[0])).astype(int)
    #find out which cluster expressed the most for each gene
    gene_value_idx = np.argmax(gene_mean,axis = 1)
    #compute DE score for each gene
    for i in range(num_cluster):
        diff = abs( np.matmul(gene_mean[:,i].reshape((num_gene,1)),np.ones((1,num_cluster))) - gene_mean)
        gene_DE_score[:,i] = np.sum( diff, axis = 1)

    #top k for each cluster based on DE score
    gclusters = []
    gscore = []
    for i in range(num_cluster):
        zz_idx = np.where(gene_value_idx == i ) # find genes' DE score in cluster i
        zz_DEscore = gene_DE_score[zz_idx,i].flatten() * (-1)
        zzvalue = np.sort(zz_DEscore, axis = 0) * (-1)
        zz1 = np.argsort(zz_DEscore, axis = 0).astype(int)
        gene_idxv = np.append(gene_idxv, zz_idx[0][zz1[0:topn]] ).astype(int)
        gclusters = np.append(gclusters, i*np.ones((topn,1)) )
        gscore = np.append(gscore, zzvalue[0:topn] )
    #GL500 = np.array([gene_idxv,gclusters,gscore]).transpose()

    #datav number of cluster * top n x number of cells
    #column vector, for each cell, the top 10 genes for each cluster that
    #differentiats the most
    gene_labels_save = gene_labels[gene_idxv]
    gene_labels_save = np.reshape(gene_labels_save, [num_cluster, topn])
    cluster_labels_save = ['C_{}'.format(i+1) for i in range(num_cluster)]
    genes_df = pd.DataFrame(gene_labels_save, index=cluster_labels_save)

    #save fig
    _, idx = np.unique(gene_idxv, return_index=True)
    gene_idxv_u = gene_idxv[np.sort(idx)]
    datav = data[gene_idxv_u,:][:,cluster_order]
    gene_labels = gene_labels[gene_idxv_u]
    gene_len = gene_labels.shape[0]
    #center is num_cluster * top n column vector, each value corresponds the average expression
    #for a gene across all num_cells
    center = np.mean(datav, axis = 1)
    #standard deviation
    scale = np.std(datav, axis = 1)
    #Check for zeros and set them to 1 so not to scale them.
    scale_ind = np.where(scale == 0)
    scale[scale_ind] = 1
    #(data - mean)/scale deviation
    sdata = np.divide(np.subtract(datav,center.reshape((gene_len,1))),scale.reshape((gene_len,1)))
    thresh = 3
    fig = plot_marker_genes(sdata, gene_labels, num_cells_inC, thresh, save_fig)
    return gene_df, fig

def plot_marker_genes(sdata, gene_labels, num_cells_inC, thresh, save_fig):
    num_cluster = len(num_cells_inC)
    num_cell = sdata.shape[1]
    gene_len = sdata.shape[0]
    #plot color map
    fig = plt.figure(figsize=(16,16))
    plt.imshow(sdata,cmap='RdBu_r',vmin=-thresh, vmax=thresh, aspect='auto')
    #plt.colorbar()
    xtkval = np.cumsum(num_cells_inC)
    xtkval1 = np.zeros(num_cluster)
    xtllab = []
    for i in range(num_cluster):
        #plt.axhline(y=i*topn, color='w')
        if i == 0:
            xtkval1[i] = 0.5*num_cells_inC[i]
            xtllab = np.append(xtllab, "C1")
            plt.axhline(xmin=0, xmax=xtkval[i]/num_cell, y=gene_len, linewidth=4, color='C'+str(i))
        else:
            xtkval1[i] = 0.5*num_cells_inC[i] +  xtkval[i-1]
            xtllab = np.append(xtllab, "C"+str(i+1))
            plt.axhline(xmin=xtkval[i-1]/num_cell, xmax=xtkval[i]/num_cell, y=gene_len,linewidth=4, color='C'+str(i%9))
    plt.xticks(xtkval1,xtllab)
    if gene_labels is None:
        plt.yticks(np.arange(gene_len))
    else:
        plt.yticks(np.arange(gene_len),gene_labels.flatten())
    return fig
